# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from pyimpspec import Circuit, ParsingError
from typing import Callable, Dict, List, Optional
from numpy import array, ndarray
import pyimpspec
import dearpygui.dearpygui as dpg
from deareis.gui.plots import Nyquist
import deareis.themes as themes
from deareis.utility import calculate_window_position_dimensions, format_number
from deareis.tooltips import attach_tooltip
import deareis.tooltips as tooltips
from deareis.gui.circuit_editor import CircuitEditor
from deareis.signals import Signal
import deareis.signals as signals
from deareis.data import DataSet


class SubtractImpedance:
    def __init__(self, data: DataSet, data_sets: List[DataSet], callback: Callable):
        assert type(data) is DataSet, data
        assert type(data_sets) is list and all(
            map(lambda _: type(_) is DataSet, data_sets)
        ), data_sets
        self.data: DataSet = data
        self.data_sets: List[DataSet] = [
            _ for _ in sorted(data_sets, key=lambda _: _.get_label()) if _ != data
        ]
        self.labels: List[str] = list(map(lambda _: _.get_label(), self.data_sets))
        self.preview_data: DataSet = DataSet.from_dict(data.to_dict())
        self.callback: Callable = callback
        self.options: List[str] = ["Constant:", " Circuit:"]
        if len(data_sets) > 0:
            self.options.append("Spectrum:")
        self.circuit_editor_window: int = -1
        x: int
        y: int
        w: int
        h: int
        x, y, w, h = calculate_window_position_dimensions()
        self.window: int = dpg.generate_uuid()
        with dpg.window(
            label="Subtract impedance",
            modal=True,
            pos=(
                x,
                y,
            ),
            width=w,
            height=h,
            tag=self.window,
            on_close=self.close,
        ):
            self.preview_window: int = dpg.generate_uuid()
            with dpg.child_window(border=False, tag=self.preview_window):
                with dpg.child_window(
                    width=-1, height=82 if len(self.data_sets) > 0 else 80
                ):
                    self.radio_buttons: int = dpg.generate_uuid()
                    with dpg.group(horizontal=True):
                        dpg.add_radio_button(
                            items=self.options,
                            default_value=self.options[0],
                            callback=self.select_option,
                            tag=self.radio_buttons,
                        )
                        with dpg.group():
                            self.constant_group: int = dpg.generate_uuid()
                            with dpg.group(horizontal=True, tag=self.constant_group):
                                dpg.add_text("Z' = ")
                                self.constant_real: int = dpg.generate_uuid()
                                dpg.add_input_float(
                                    label="ohm,",
                                    default_value=0.0,
                                    step=0.0,
                                    on_enter=True,
                                    width=100,
                                    tag=self.constant_real,
                                    callback=self.update_preview,
                                )
                                dpg.add_text('-Z" = ')
                                self.constant_imag: int = dpg.generate_uuid()
                                dpg.add_input_float(
                                    label="ohm",
                                    default_value=0.0,
                                    step=0.0,
                                    on_enter=True,
                                    width=100,
                                    tag=self.constant_imag,
                                    callback=self.update_preview,
                                )
                            self.circuit_group: int = dpg.generate_uuid()
                            with dpg.group(horizontal=True, tag=self.circuit_group):
                                self.circuit_cdc: int = dpg.generate_uuid()
                                dpg.add_input_text(
                                    hint="Input CDC",
                                    on_enter=True,
                                    width=314,
                                    tag=self.circuit_cdc,
                                    callback=self.update_preview,
                                )
                                dpg.add_button(
                                    label="Edit",
                                    callback=self.edit_circuit,
                                )
                                attach_tooltip(tooltips.general.open_circuit_editor)
                            if len(self.data_sets) > 0:
                                self.spectrum_group: int = dpg.generate_uuid()
                                with dpg.group(
                                    horizontal=True, tag=self.spectrum_group
                                ):
                                    self.spectrum_combo: int = dpg.generate_uuid()
                                    dpg.add_combo(
                                        items=self.labels,
                                        default_value=self.labels[0],
                                        width=358,
                                        tag=self.spectrum_combo,
                                        callback=self.update_preview,
                                    )
                self.nyquist_plot: Nyquist = Nyquist(width=-1, height=-24)
                dpg.add_button(
                    label="Accept",
                    callback=self.accept,
                )
            self.circuit_editor_window = dpg.generate_uuid()
            with dpg.child_window(
                border=False,
                show=False,
                tag=self.circuit_editor_window,
            ):
                self.circuit_editor: CircuitEditor = CircuitEditor(
                    window=self.circuit_editor_window,
                    callback=self.accept_circuit,
                )
        self.key_handler: int = dpg.generate_uuid()

        def keybinding_handler(sender: int, key: int, callback: Callable):
            if not dpg.is_item_visible(self.constant_real):
                return
            if (
                dpg.is_item_focused(self.constant_real)
                or dpg.is_item_focused(self.constant_imag)
                or dpg.is_item_focused(self.circuit_cdc)
            ):
                return
            callback()

        with dpg.handler_registry(tag=self.key_handler):
            dpg.add_key_release_handler(
                key=dpg.mvKey_Escape,
                callback=keybinding_handler,
                user_data=self.close,
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Return,
                callback=keybinding_handler,
                user_data=self.accept,
            )
        self.select_option(self.radio_buttons, self.options[0])

    def close(self):
        self.circuit_editor.hide()
        dpg.hide_item(self.window)
        dpg.delete_item(self.window)
        dpg.delete_item(self.key_handler)
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)

    def accept(self):
        if (
            self.circuit_editor_window > 0
            and dpg.does_item_exist(self.circuit_editor_window)
            and dpg.is_item_shown(self.circuit_editor_window)
        ):
            return
        self.close()
        self.callback(DataSet.from_dict(self.preview_data.to_dict()))

    def select_option(self, sender: int, value: str):
        item_type: str

        def disable_group(group: int):
            for item in dpg.get_item_children(group, slot=1):
                item_type = dpg.get_item_type(item)
                if item_type.endswith("mvText") or item_type.endswith("mvTooltip"):
                    continue
                dpg.disable_item(item)

        def enable_group(group: int):
            for item in dpg.get_item_children(group, slot=1):
                item_type = dpg.get_item_type(item)
                if item_type.endswith("mvText") or item_type.endswith("mvTooltip"):
                    continue
                dpg.enable_item(item)

        index: int = self.options.index(value)
        if index == 0:
            enable_group(self.constant_group)
            disable_group(self.circuit_group)
            disable_group(self.spectrum_group)
        elif index == 1:
            disable_group(self.constant_group)
            enable_group(self.circuit_group)
            disable_group(self.spectrum_group)
        elif index == 2:
            disable_group(self.constant_group)
            disable_group(self.circuit_group)
            enable_group(self.spectrum_group)
        else:
            raise Exception(f"Unsupported option: {value}")
        self.update_preview()

    def update_preview(self):
        index: int = self.options.index(dpg.get_value(self.radio_buttons))
        f: ndarray = self.data.get_frequency(masked=None)
        Z: ndarray = self.data.get_impedance(masked=None)
        if index == 0:
            Z_const: complex = complex(
                dpg.get_value(self.constant_real), -dpg.get_value(self.constant_imag)
            )
            Z = Z - Z_const
        elif index == 1:
            try:
                circuit: Circuit = pyimpspec.string_to_circuit(
                    dpg.get_value(self.circuit_cdc)
                )
            except Exception as e:
                # print(e)
                return
            Z = Z - circuit.impedances(f)
        elif index == 2:
            spectrum: DataSet
            spectrum = self.data_sets[
                self.labels.index(dpg.get_value(self.spectrum_combo))
            ]
            Z = Z - spectrum.get_impedance(masked=None)
        else:
            raise Exception(f"Unsupported option!")
        self.preview_data._impedance = Z
        self.update_plot()

    def update_plot(self):
        self.nyquist_plot.clear()
        real: ndarray
        imag: ndarray
        real, imag = self.data.get_nyquist_data(masked=None)
        self.nyquist_plot.plot(
            real=real,
            imaginary=imag,
            label="Before",
            theme=themes.nyquist.data,
            show_label=False,
        )
        self.nyquist_plot.plot(
            real=real,
            imaginary=imag,
            label="Before",
            line=True,
            theme=themes.nyquist.data,
        )
        real, imag = self.preview_data.get_nyquist_data(masked=None)
        self.nyquist_plot.plot(
            real=real,
            imaginary=imag,
            label="After",
            theme=themes.bode.phase_data,
            show_label=False,
        )
        self.nyquist_plot.plot(
            real=real,
            imaginary=imag,
            label="After",
            line=True,
            theme=themes.bode.phase_data,
        )
        self.nyquist_plot.queue_limits_adjustment()

    def edit_circuit(self):
        dpg.hide_item(self.preview_window)
        circuit: Optional[Circuit] = None
        try:
            circuit = pyimpspec.string_to_circuit(
                dpg.get_value(self.circuit_cdc) or "[]"
            )
        except ParsingError:
            pass
        self.circuit_editor.show(circuit)

    def accept_circuit(self, circuit: Optional[Circuit]):
        self.circuit_editor.hide()
        dpg.show_item(self.preview_window)
        self.update_cdc(circuit)

    def update_cdc(self, circuit: Optional[Circuit]):
        if circuit is not None:
            for element in circuit.get_elements():
                element.set_label("")
                for param in element.get_parameters():
                    element.set_fixed(param, True)
        assert dpg.does_item_exist(self.circuit_cdc)
        dpg.set_value(
            self.circuit_cdc, circuit.to_string(6) if circuit is not None else ""
        )
        dpg.show_item(self.preview_window)
        dpg.split_frame(delay=33)
        self.update_preview()



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from deareis.data.kramers_kronig import TestSettings
from deareis.tooltips import attach_tooltip
from deareis.utility import calculate_window_position_dimensions
from numpy import array, floor, log10 as log, ndarray
from pyimpspec import KramersKronigResult
from typing import Callable, Dict, List, Tuple
import deareis.themes as themes
import deareis.tooltips as tooltips
import dearpygui.dearpygui as dpg
from deareis.signals import Signal
import deareis.signals as signals
from deareis.data import DataSet
import pyimpspec
from deareis.gui.plots import (
    Bode,
    MuXps,
    Nyquist,
    Residuals,
)


class ExploratoryResults:
    def __init__(
        self,
        data: DataSet,
        results: List[KramersKronigResult],
        settings: TestSettings,
        num_RCs: ndarray,
        callback: Callable,
        state,
    ):
        assert type(data) is DataSet
        assert type(results) is list and all(
            map(lambda _: type(_) is KramersKronigResult, results)
        )
        assert type(settings) is TestSettings
        assert type(num_RCs) is ndarray
        self.state = state
        dpg.split_frame(delay=33)
        self.callback: Callable = callback
        self.window: int = dpg.generate_uuid()
        self.result_combo: int = dpg.generate_uuid()
        self.accept_button: int = dpg.generate_uuid()
        self.mu_xps_plot: MuXps = None
        self.residuals_plot: Residuals = None
        self.nyquist_plot: Nyquist = None
        self.bode_plot: Bode = None
        self.key_handler: int = dpg.generate_uuid()
        self._assemble()
        self._setup_keybindings()
        self.data: DataSet = data
        results.sort(key=lambda _: _.num_RC)
        self.results: List[KramersKronigResult] = results
        self.settings: TestSettings = settings
        self.num_RCs: ndarray = num_RCs
        self.mu_crit: float = settings.mu_criterion
        default_result: KramersKronigResult = pyimpspec.score_test_results(
            results, self.mu_crit
        )[0][1]
        default_label: str = ""
        self.label_to_result: Dict[str, KramersKronigResult] = {}
        result: KramersKronigResult
        for result in results:
            label: str = (
                f"{result.num_RC}: µ = {result.mu:.3f}, "
                + f"log X² (ps.) = {log(result.pseudo_chisqr):.3f}"
            )
            if result == default_result:
                label += " *"
                default_label = label
            self.label_to_result[label] = result
        assert default_label != ""
        self.labels: List[str] = list(self.label_to_result.keys())
        self.result_index: int = self.labels.index(default_label)
        dpg.configure_item(
            self.result_combo,
            items=self.labels,
            default_value=default_label,
        )
        self.mu: ndarray = array(list(map(lambda _: _.mu, results)))
        self.Xps: ndarray = log(array(list(map(lambda _: _.pseudo_chisqr, results))))
        self.plot(default_label)
        signals.register(Signal.VIEWPORT_RESIZED, self.resize)

    def _setup_keybindings(self):
        with dpg.handler_registry(tag=self.key_handler):
            dpg.add_key_release_handler(
                key=dpg.mvKey_Escape,
                callback=self.close,
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Return,
                callback=lambda: self.accept(
                    dpg.get_item_user_data(self.accept_button)
                ),
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Prior,
                callback=lambda: self.plot(
                    self.labels[(self.result_index - 1) % len(self.labels)]
                ),
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Next,
                callback=lambda: self.plot(
                    self.labels[(self.result_index + 1) % len(self.labels)]
                ),
            )

    def _assemble(self):
        x: int
        y: int
        w: int
        h: int
        x, y, w, h = calculate_window_position_dimensions()
        with dpg.window(
            label="Exploratory test results",
            modal=True,
            pos=(
                x,
                y,
            ),
            width=w,
            height=h,
            no_resize=True,
            tag=self.window,
            on_close=self.close,
        ):
            with dpg.group(horizontal=True):
                dpg.add_text("Num. RC circuits")
                attach_tooltip(tooltips.kramers_kronig.exploratory_result)
                dpg.add_combo(
                    width=-100,
                    tag=self.result_combo,
                )
                dpg.add_button(
                    label="Accept",
                    width=-1,
                    tag=self.accept_button,
                )
            with dpg.child_window(
                border=True,
                width=-1,
                height=-1,
            ):
                with dpg.group(horizontal=True):
                    self.mu_xps_plot = MuXps()
                    self.residuals_plot = Residuals()
                with dpg.group(horizontal=True):
                    self.nyquist_plot = Nyquist()
                    self.bode_plot = Bode()
        dpg.set_item_callback(self.result_combo, lambda s, a: self.plot(a))
        dpg.set_item_callback(
            self.accept_button,
            lambda s, a, u: self.accept(u),
        )
        self.resize(1, 1)

    def plot(self, label: str):
        self.result_index = self.labels.index(label)
        dpg.set_value(self.result_combo, label)
        # Clear plots
        self.mu_xps_plot.clear()
        self.residuals_plot.clear()
        self.nyquist_plot.clear()
        self.bode_plot.clear()
        # Retrieve the chosen result
        result: KramersKronigResult = self.label_to_result[label]
        dpg.set_item_user_data(self.accept_button, result)
        # Mu-Xps vs num RC
        self.mu_xps_plot.plot(
            num_RCs=self.num_RCs,
            mu=self.mu,
            Xps=self.Xps,
            mu_criterion=self.mu_crit,
            num_RC=result.num_RC,
        )
        # Residuals
        freq: ndarray
        real: ndarray
        imag: ndarray
        freq, real, imag = result.get_residual_data()
        self.residuals_plot.plot(
            frequency=freq,
            real=real,
            imaginary=imag,
        )
        # Data and fit
        # - Nyquist
        real, imag = self.data.get_nyquist_data()
        self.nyquist_plot.plot(
            real=real,
            imaginary=imag,
            label="Data",
            theme=themes.nyquist.data,
        )
        real, imag = result.get_nyquist_data(
            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
        )
        self.nyquist_plot.plot(
            real=real,
            imaginary=imag,
            label="Fit",
            show_label=False,
            line=True,
            theme=themes.nyquist.simulation,
        )
        real, imag = result.get_nyquist_data()
        self.nyquist_plot.plot(
            real=real,
            imaginary=imag,
            label="Fit",
            theme=themes.nyquist.simulation,
        )
        # Bode
        mag: ndarray
        phase: ndarray
        freq, mag, phase = self.data.get_bode_data()
        self.bode_plot.plot(
            frequency=freq,
            magnitude=mag,
            phase=phase,
            labels=(
                "|Z| (d)",
                "phi (d)",
            ),
            themes=(
                themes.bode.magnitude_data,
                themes.bode.phase_data,
            ),
        )
        freq, mag, phase = result.get_bode_data(
            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
        )
        self.bode_plot.plot(
            frequency=freq,
            magnitude=mag,
            phase=phase,
            labels=("|Z| (f)", "phi (f)"),
            show_labels=False,
            line=True,
            themes=(
                themes.bode.magnitude_simulation,
                themes.bode.phase_simulation,
            ),
        )
        freq, mag, phase = result.get_bode_data()
        self.bode_plot.plot(
            frequency=freq,
            magnitude=mag,
            phase=phase,
            labels=("|Z| (f)", "phi (f)"),
            themes=(
                themes.bode.magnitude_simulation,
                themes.bode.phase_simulation,
            ),
        )
        dpg.split_frame()
        self.mu_xps_plot.queue_limits_adjustment()
        self.residuals_plot.queue_limits_adjustment()
        self.nyquist_plot.queue_limits_adjustment()
        self.bode_plot.queue_limits_adjustment()

    def resize(self, width: int, height: int):
        x: int
        y: int
        w: int
        h: int
        x, y, w, h = calculate_window_position_dimensions()
        dpg.configure_item(
            self.window,
            pos=(
                x,
                y,
            ),
            width=w,
            height=h,
        )
        dpg.split_frame()
        width, height = dpg.get_item_rect_size(self.window)
        width -= 56
        height -= 80
        width = int(floor(width / 2))
        height = int(floor(height / 2))
        self.mu_xps_plot.resize(width, height)
        self.residuals_plot.resize(width, height)
        self.nyquist_plot.resize(width, height)
        self.bode_plot.resize(width, height)

    def close(self):
        dpg.hide_item(self.window)
        dpg.delete_item(self.window)
        dpg.delete_item(self.key_handler)
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)
        signals.unregister(Signal.VIEWPORT_RESIZED, self.resize)

    def accept(self, result: KramersKronigResult):
        self.callback(self.data, result, self.settings)
        self.close()
